@@ -1,12 +1,12 @@
 ## R routines for gam fitting with calculation of derivatives w.r.t. sp.s
-## (c) Simon Wood 2004,2005
+## (c) Simon Wood 2004,2005,2006
 
 gam.fit2 <- function (x, y, sp, S=list(),rS=list(),off, H=NULL, weights =
 rep(1, nobs), start = NULL, etastart = NULL, 
     mustart = NULL, offset = rep(0, nobs), family = gaussian(), 
     control = gam.control(), intercept =
     TRUE,deriv=TRUE,gamma=1,scale=1,pearson=FALSE,
-    printWarn=TRUE) 
+    printWarn=TRUE,...) 
 ## deriv, sp, S, H added to arg list. 
 ## need to modify family before call.
 {   x <- as.matrix(x)
@@ -15,7 +15,7 @@ rep(1, nobs), start = NULL, etastart = NULL,
         rownames(y)
     else names(y)
     conv <- FALSE
-    nobs <- NROW(y)
+    n <- nobs <- NROW(y) ## n is just to keep codetools happy
     nvars <- ncol(x)
     EMPTY <- nvars == 0
     if (is.null(weights)) 
@@ -215,9 +215,9 @@ rep(1, nobs), start = NULL, etastart = NULL,
 
             if (iter>1&&pdev>old.pdev) { ## solution diverging
               ii <- 1
-             # while (pdev - old.pdev> (0.1 + abs(old.pdev))* control$epsilon*.9)
-            while (pdev -old.pdev > (.1+abs(old.pdev))*.2)  
-            {
+           ## while (pdev -old.pdev > (.1+abs(old.pdev))*.2)  
+            while (pdev -old.pdev > (.1+abs(old.pdev))*.Machine$double.eps*10)  
+             { ## step halve until pdev <= old.pdev
                 if (ii > 200) 
                    stop("inner loop 3; can't correct step size")
                 ii <- ii + 1
@@ -347,7 +347,7 @@ rep(1, nobs), start = NULL, etastart = NULL,
 
 
 
-gam2derivative <- function(lsp,args)
+gam2derivative <- function(lsp,args,...)
 ## Performs IRLS GAM fitting for smoothing parameters given in lsp 
 ## and returns the derivatives of the GCV or UBRE score w.r.t the 
 ## smoothing parameters for the model.
@@ -356,13 +356,13 @@ gam2derivative <- function(lsp,args)
 { b<-gam.fit2(x=args$X, y=args$y, sp=lsp, S=args$S,rS=args$rS,off=args$off, H=args$H,
      offset = args$offset,family = args$family,weights=args$w,deriv=TRUE,
      control=args$control,gamma=args$gamma,scale=args$scale,pearson=args$pearson,
-     printWarn=FALSE)
+     printWarn=FALSE,...)
   if (args$scoreType == "GCV") ret <- b$GCV1 else ret <- b$UBRE1
   ret
 }
 
 
-gam2objective <- function(lsp,args,printWarn=FALSE)
+gam2objective <- function(lsp,args,printWarn=FALSE,...)
 ## Performs IRLS GAM fitting for smoothing parameters given in lsp 
 ## and returns the GCV or UBRE score for the model.
 ## args is a list containing the arguments for gam.fit2
@@ -371,13 +371,13 @@ gam2objective <- function(lsp,args,printWarn=FALSE)
   b<-gam.fit2(x=args$X, y=args$y, sp=lsp, S=args$S,rS=args$rS,off=args$off, H=args$H,
      offset = args$offset,family = args$family,weights=args$w,deriv=FALSE,
      control=args$control,gamma=args$gamma,scale=args$scale,pearson=args$pearson,
-     printWarn=printWarn)
+     printWarn=printWarn,...)
   if (args$scoreType == "GCV") ret <- b$GCV else ret <- b$UBRE
   attr(ret,"full.fit") <- b
   ret
 }
 
-gam3objective <- function(lsp,args)
+gam3objective <- function(lsp,args,...)
 ## Performs IRLS GAM fitting for smoothing parameters given in lsp 
 ## and returns the GCV or UBRE score for the model.
 ## args is a list containing the arguments for gam.fit2
@@ -386,7 +386,7 @@ gam3objective <- function(lsp,args)
   b<-gam.fit2(x=args$X, y=args$y, sp=lsp, S=args$S,rS=args$rS,off=args$off, H=args$H,
      offset = args$offset,family = args$family,weights=args$w,deriv=TRUE,
      control=args$control,gamma=args$gamma,scale=args$scale,pearson=args$pearson,
-     printWarn=FALSE)
+     printWarn=FALSE,...)
   if (args$scoreType == "GCV") ret <- b$GCV else ret <- b$UBRE
   attr(ret,"full.fit") <- b
   if (args$scoreType == "GCV") at <- b$GCV1 else at <- b$UBRE1
@@ -394,34 +394,44 @@ gam3objective <- function(lsp,args)
   ret
 }
 
+
+
 fix.family.link<-function(fam)
 # adds d2link the second derivative of the link function w.r.t. mu
-# to the family supplied.
-# All d2link functions have been checked numerically. 
+# to the family supplied, as well as a 3rd derivative function 
+# d3link...
+# All d2link and d3link functions have been checked numerically. 
 { if (!inherits(fam,"family")) stop("fam not a family object")
   if (!is.null(fam$d2link)) return(fam) 
   link <- fam$link
   if (length(link)>1) if (fam$family=="quasi") # then it's a power link
-  { lambda <- log(fam$linkfun(exp(1)))
-    if (lambda<=0) fam$d2link <- function(mu) -1/mu^2
-    else fam$d2link <- function(mu) lambda*(lambda-1)*mu^(lambda-2)
+  { lambda <- log(fam$linkfun(exp(1))) ## the power, if > 0
+    if (lambda<=0) { fam$d2link <- function(mu) -1/mu^2
+      fam$d3link <- function(mu) 2/mu^3
+    }
+    else { fam$d2link <- function(mu) lambda*(lambda-1)*mu^(lambda-2)
+      fam$d3link <- function(mu) (lambda-2)*(lambda-1)*lambda*mu^(lambda-3)
+    }
     return(fam)
   } else stop("unrecognized (vector?) link")
 
   if (link=="identity") {
-    fam$d2link <- function(mu) rep.int(0,length(mu))
+    fam$d3link <- fam$d2link <- function(mu) rep.int(0,length(mu))
     return(fam)
   } 
   if (link == "log") {
     fam$d2link <- function(mu) -1/mu^2
+    fam$d3link <- function(mu) 2/mu^3
     return(fam)
   }
   if (link == "inverse") {
     fam$d2link <- function(mu) 2/mu^3
+    fam$d3link <- function(mu) {mu <- mu*mu;-6/(mu*mu)}
     return(fam)
   }
   if (link == "logit") {
     fam$d2link <- function(mu) 1/(1 - mu)^2 - 1/mu^2
+    fam$d3link <- function(mu) 2/(1 - mu)^3 + 2/mu^3
     return(fam)
   }
   if (link == "probit") {
@@ -429,21 +439,43 @@ fix.family.link<-function(fam)
       eta <- fam$linkfun(mu)
       eta/fam$mu.eta(eta)^2
     }
+    fam$d3link <- function(mu) {
+      eta <-  fam$linkfun(mu)
+      (1 + 2*eta^2)/fam$mu.eta(eta)^3
+    }
     return(fam)
   }
   if (link == "cloglog") {
-    fam$d2link <- function(mu) {
-      -(1/(1 - mu)^2/log(1 - mu) + 
-       1/(1 - mu) * (1/(1 - mu))/log(1 - mu)^2)
+    fam$d2link <- function(mu) { l1m <- log(1-mu)
+      -1/((1 - mu)^2*l1m) *(1+ 1/l1m)
+    }
+    fam$d3link <- function(mu) { l1m <- log(1-mu)
+       mu3 <- (1-mu)^3
+      -1/(mu3 * l1m^3) -(1 + 2 * l1m)/
+       (mu3 * l1m^2) * (1 + 1/l1m)
     }
     return(fam)
   }
   if (link == "sqrt") {
     fam$d2link <- function(mu) -.25 * mu^-1.5
+    fam$d3link <- function(mu) .375 * mu^-2.5
+    return(fam)
+  }
+  if (link == "cauchit") {
+    fam$d2link <- function(mu) { 
+     eta <- fam$linkfun(mu)
+     2*pi*pi*eta*(1+eta*eta)
+    }
+    fam$d3link <- function(mu) { 
+     eta <- fam$linkfun(mu)
+     eta2 <- eta*eta
+     2*pi*pi*pi*(1+3*eta2)*(1+eta2)
+    }
     return(fam)
   }
   if (link == "1/mu^2") {
     fam$d2link <- function(mu) 6 * mu^-4
+    fam$d3link <- function(mu) -24* mu^-5
     return(fam)
   }
   stop("link not recognised")
@@ -452,24 +484,28 @@ fix.family.link<-function(fam)
 
 fix.family.var<-function(fam)
 # adds dvar the derivative of the variance function w.r.t. mu
-# to the family supplied.
+# to the family supplied, as well as d2var the 2nd derivative of 
+# the variance function w.r.t. the mean. (All checked numerically). 
 { if (!inherits(fam,"family")) stop("fam not a family object")
   if (!is.null(fam$dvar)) return(fam) 
   family <- fam$family
   if (family=="gaussian") {
-    fam$dvar <- function(mu) rep.int(0,length(mu))
+    fam$d2var <- fam$dvar <- function(mu) rep.int(0,length(mu))
     return(fam)
   } 
   if (family=="poisson"||family=="quasipoisson") {
     fam$dvar <- function(mu) rep.int(1,length(mu))
+    fam$d2var <- function(mu) rep.int(0,length(mu))
     return(fam)
   } 
   if (family=="binomial"||family=="quasibinomial") {
     fam$dvar <- function(mu) 1-2*mu
+    fam$d2var <- function(mu) rep.int(-2,length(mu))
     return(fam)
   }
   if (family=="Gamma") {
     fam$dvar <- function(mu) 2*mu
+    fam$d2var <- function(mu) rep.int(2,length(mu))
     return(fam)
   }
   if (family=="quasi") {
@@ -481,15 +517,25 @@ fix.family.var<-function(fam)
        "mu^3" = function(mu) 3*mu^2           
     )
     if (is.null(fam$dvar)) stop("variance function not recognized for quasi")
+    fam$d2var <- switch(fam$varfun,
+       constant = function(mu) rep.int(0,length(mu)),
+       "mu(1-mu)" = function(mu) rep.int(-2,length(mu)),
+       mu = function(mu) rep.int(0,length(mu)),
+       "mu^2" = function(mu) rep.int(2,length(mu)),
+       "mu^3" = function(mu) 6*mu           
+    )
     return(fam)
   }
   if (family=="inverse.gaussian") {
     fam$dvar <- function(mu) 3*mu^2
+    fam$d2var <- function(mu) 6*mu
     return(fam)
   }
   stop("family not recognised")
 }
 
+
+
 totalPenalty <- function(S,H,off,theta,p)
 { if (is.null(H)) St <- matrix(0,p,p)
   else { St <- H; 
